/*******************************************************************************
*    Autocomplete                                                               *
*                                                                              *
*    Copyright (C) 2016 Luca Bertoncello                                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

var Autocomplete = (function() {
  var _AUTOCOMPLETE_VERSION = '1.0-20160911';
  var _AUTOCOMPLETE_DATA = {};

  /**
   * Returns the version number of the script
   */
  var version = function()
  {
    return _AUTOCOMPLETE_VERSION;
  };

  /**
   * Send an AJAX-Request
   * 
   * @param string url             The URL of the AJAX-Script
   * @param string data            The data to be sent
   * @param function cbFunction    A callback function to be called on the answer
   * @param array cbParams         Parameter to the callback function
   *
   * The callback function must be defined as follow:
   * 
   *    function foo(response, status, params)
   * 
   * response is the response of the AJAX-Script (as plain text)
   * status is the HTTP status code of the query
   * params is an array with optional parameters (@see cbParams)
   */
  var sendAJAXRequest = function(url, data, cbFunction, cbParams)
  {
    var req;

    req = new XMLHttpRequest();

    req.onreadystatechange = function()
    {
      if(req.readyState == 4 && req.status >= 200 && req.status < 300)
        cbFunction(req.responseText, req.status, cbParams);
    };

    req.open("POST", url, true);
    req.setRequestHeader("Content-type", "application/x-www-form-urlencoded");
    req.send(data);
  };

  /**
   * General function to retrieve the properties of an element,
   * analog to getBoundingClientRect(), which does not work on Opera 9.27
   * 
   * @param element elem           The element
   * @returns array                Array with top, left, width and height
   */
  var getElementRect = function(elem)
  {
    var top = 0, left = 0;
    var cs = document.defaultView.getComputedStyle(elem, null);
    var width = cs.width.replace('px', '');
    var height = cs.height.replace('px', '');

    while(elem)
    {
      top = top + parseInt(elem.offsetTop);
      left = left + parseInt(elem.offsetLeft);
      elem = elem.offsetParent;
    }

    return { 'top': top, 'left': left, 'width': parseInt(width), 'height': parseInt(height) };
  };

  /**
   * Close the droplist with the results
   */
  var closeDroplist = function()
  {
  //Close dropdown
    var elems = document.getElementsByTagName('div');
    for(var l = 0; l < elems.length; l++)
    {
      if(elems[l].className == 'ac_dropdown')
        document.body.removeChild(elems[l]);
    }
  // Remove class from input field
    elems = document.getElementsByTagName('input');
    for(var l = 0; l < elems.length; l++)
      elems[l].className = elems[l].className.replace('ac_maininput', '');
  };

  /**
   * Returns the extra parameters for the onXXX functions
   * 
   * @param string elementID                The element ID of the text-field to set the autocomplete to
   * @returns mixed                         The extra parameters for the functions
   */
  var getExtraJSParams = function(elementID)
  {
    if(_AUTOCOMPLETE_DATA[elementID]['jsParams'] != undefined)
      return _AUTOCOMPLETE_DATA[elementID]['jsParams'];
    return null;
  };

  /**
   * Define the autocomplete for the given element
   * 
   * @param array params                    The parameters as associative array. Keys:
   *                   string elementID     The element ID of the text-field to set the autocomplete to
   *                   string ajaxURL       The URL of the backend for the autocomplete
   *                   string searchKey     The parameter of the AJAX function that specify the search element
   *                   array ajaxParams     Optional parameters for the AJAX function
   *                                        This is an associative array of the optional parameters.
   *                                        A parameter can be a value or a function, that will be evaluated at
   *                                        the time of the AJAX-Request
   *                   mixed jsParams       Optional parameter(s) for the onXXX functions.
   *                                        This can be an object, string, number, array or even a function.
   *                                        The function will get these values using the function
   *                                                getExtraJSParams(elementID)
   *                   int minLength        The minimum lenght of the string to start the autocomplete (default: 2)
   *                   bool deleteOnClick   Deletes the value of the element on mouse click (default: false)
   *                   function onSelect    Optional function to be called if an item was selected.
   *                                        It receives the choosen value as parameter and an optional associative
   *                                        array of extra data, if a "lineParser" was defined
   *                   function lineParser  Optional function to handle complex answers. It receives as parameter
   *                                        the current line from AJAX and should returns an associative array
   *                                        with key "value" for the value of the line and optional other free
   *                                        defined keys to be passed as extra data to the line.
   *                   array onKey          Optional list of special keys (eg: TAB) with the function to be called.
   *                                        This is an associative array with key as numerical keyCode of the key and
   *                                        value as associative array. The keys are "close" (boolean) to allow closing
   *                                        of the droplist on key pressed, and "func" as the function to be called.
   *                                        The called function receives the current text value, the current
   *                                        selected entry (if any!) as parameter and optionally the array of extra data,
   *                                        if a "lineParser" was defined
   *                   int scrollAfter      Dropdown show a scrollbar if the list is longer than given elements (default: no limit)
   *                   string cbFunction    Optional callback function to handle the request
   *                   array cbParams       Optional parameters for the callback function
   * 
   * If no callback function will be given, a dropdown list will be displayed.
   * Otherwise the callback function will be called
   * Function must have this prototype:
   * 
   *    function foo(data, response, cbParams)
   *
   * where "data" is the value of the searched text (can be empty!!),
   * "response" is the response of the AJAX-Script and "cbParams" the parameter
   * given to setAutocomplete()
   * 
   * The AJAX-function should return a JSON-encoded numeric array of elements if no callback
   * function was defined or complex data in other case (handled in custom function).
   */
  var setAutocomplete = function(params)
  {
    var elementID, minLength, ajaxURL, searchKey, ajaxParams, cbFunction, cbParams, l, scrollAfter, onKey;

    elementID = params.elementID;
    ajaxURL = params.ajaxURL;
    searchKey = params.searchKey;
    ajaxParams = params.ajaxParams;
    minLength = params.minLength || 2;
    cbFunction = params.cbFunction;
    cbParams = params.cbParams;
    scrollAfter = params.scrollAfter || false;
    onKey = params.onKey || false;
    _AUTOCOMPLETE_DATA[elementID] = { onSelect      : params.onSelect,
                                      lineParser    : params.lineParser,
                                      deleteOnClick : params.deleteOnClick || false,
                                      jsParams      : params.jsParams,
                                      curSelection  : -1,
                                      curValue      : ''};

    if(document.getElementById(elementID) === null)
      return;
    document.getElementById(elementID).setAttribute('autocomplete', 'off');
    if(onKey !== false)
    {
      document.getElementById(elementID).onkeypress = function(event)
      {
        var droplist = null, dlUL = null, curSel = null;
        for(var keyC in onKey)
        {
          if(event.keyCode == keyC)
          {
            var userData = {};
            curSel = null;
            if(droplist !== null)
            {
              dlUL = droplist.children[0];
              curSel = dlUL.children[_AUTOCOMPLETE_DATA[elementID]['curSelection']].getAttribute('data-value');
              if(dlUL.children[_AUTOCOMPLETE_DATA[elementID]['curSelection']].getAttribute('data-complex') !== null &&
                 dlUL.children[_AUTOCOMPLETE_DATA[elementID]['curSelection']].getAttribute('data-complex') != '')
              {
                var dataComplex = Base64.decode(dlUL.children[_AUTOCOMPLETE_DATA[elementID]['curSelection']].getAttribute('data-complex'));
                var data = dataComplex.split(',');
                for(l = 0; l < data.length; l++)
                {
                  var x = data[l].split('=');
                  userData[x[0]] = x[1];
                }
              }
            }
            if(onKey[keyC]['close'] === true)
              closeDroplist();
            if(onKey[keyC]['func'] != undefined)
              onKey[keyC]['func'](document.getElementById(elementID).value, curSel, userData);
          }
        }
      };
    }
    document.getElementById(elementID).onkeyup = function(event)
    {
      var droplist = document.getElementById('ac_' + elementID + '_dropdown');
// http://www.mediaevent.de/javascript/Extras-Javascript-Keycodes.html
      if(event.keyCode == 13) // enter
      {
        if(droplist !== null)
        {
          dlUL = droplist.children[0];
          document.getElementById(elementID).value = dlUL.children[_AUTOCOMPLETE_DATA[elementID]['curSelection']].getAttribute('data-value');
          closeDroplist();
          if(_AUTOCOMPLETE_DATA[elementID]['onSelect'] != undefined)
          {
            var userData = {};
            if(dlUL.children[_AUTOCOMPLETE_DATA[elementID]['curSelection']].hasAttribute('data-complex'))
            {
              if(dlUL.children[_AUTOCOMPLETE_DATA[elementID]['curSelection']].getAttribute('data-complex') !== null &&
                 dlUL.children[_AUTOCOMPLETE_DATA[elementID]['curSelection']].getAttribute('data-complex') != '')
              {
                var dataComplex = Base64.decode(dlUL.children[_AUTOCOMPLETE_DATA[elementID]['curSelection']].getAttribute('data-complex'));
                var data = dataComplex.split(',');
                for(l = 0; l < data.length; l++)
                {
                  var x = data[l].split('=');
                  userData[x[0]] = x[1];
                }
              }
            }
            _AUTOCOMPLETE_DATA[elementID]['onSelect'](document.getElementById(elementID).value, userData);
          }
        }
      }
      else if(event.keyCode == 33) // page up
      {
        if(droplist !== null)
        {
          dlUL = droplist.children[0];
          _AUTOCOMPLETE_DATA[elementID]['curSelection'] = 0;
          for(l = 0; l < dlUL.children.length; l++)
          {
            if(l == _AUTOCOMPLETE_DATA[elementID]['curSelection'])
            {
              dlUL.children[l].className = 'hover';
              document.getElementById(elementID).value = dlUL.children[l].getAttribute('data-value');
            }
            else
              dlUL.children[l].className = '';
          }
          dlUL.scrollTop = 0;
        }
        else
          _AUTOCOMPLETE_DATA[elementID]['curSelection'] = -1;
      }
      else if(event.keyCode == 34) // page down
      {
        if(droplist !== null)
        {
          dlUL = droplist.children[0];
          _AUTOCOMPLETE_DATA[elementID]['curSelection'] = dlUL.children.length - 1;
          for(l = 0; l < dlUL.children.length; l++)
          {
            if(l == _AUTOCOMPLETE_DATA[elementID]['curSelection'])
            {
              dlUL.children[l].className = 'hover';
              document.getElementById(elementID).value = dlUL.children[l].getAttribute('data-value');
            }
            else
              dlUL.children[l].className = '';
          }
          dlUL.scrollTop = dlUL.children.length * 20;
        }
        else
          _AUTOCOMPLETE_DATA[elementID]['curSelection'] = -1;
      }
      else if(event.keyCode == 38) // up arrow
      {
        if(droplist !== null)
        {
          dlUL = droplist.children[0];
          _AUTOCOMPLETE_DATA[elementID]['curSelection']--;
          for(l = 0; l < dlUL.children.length; l++)
          {
            if(l == _AUTOCOMPLETE_DATA[elementID]['curSelection'])
            {
              dlUL.children[l].className = 'hover';
              document.getElementById(elementID).value = dlUL.children[l].getAttribute('data-value');
            }
            else
              dlUL.children[l].className = '';
          }
          if(_AUTOCOMPLETE_DATA[elementID]['curSelection'] >= scrollAfter)
            dlUL.scrollTop -= 20;
          if(_AUTOCOMPLETE_DATA[elementID]['curSelection'] <= -1)
          {
            _AUTOCOMPLETE_DATA[elementID]['curSelection'] = dlUL.children.length;
            document.getElementById(elementID).value = _AUTOCOMPLETE_DATA[elementID]['curValue'];
            dlUL.scrollTop = _AUTOCOMPLETE_DATA[elementID]['curSelection'] * 20;
          }
        }
        else
          _AUTOCOMPLETE_DATA[elementID]['curSelection'] = -1;
      }
      else if(event.keyCode == 40) // down arrow
      {
        if(droplist !== null)
        {
          dlUL = droplist.children[0];
          _AUTOCOMPLETE_DATA[elementID]['curSelection']++;
          for(l = 0; l < dlUL.children.length; l++)
          {
            if(l == _AUTOCOMPLETE_DATA[elementID]['curSelection'])
            {
              dlUL.children[l].className = 'hover';
              document.getElementById(elementID).value = dlUL.children[l].getAttribute('data-value');
            }
            else
              dlUL.children[l].className = '';
          }
          if(_AUTOCOMPLETE_DATA[elementID]['curSelection'] >= scrollAfter)
            dlUL.scrollTop += 20;
          if(_AUTOCOMPLETE_DATA[elementID]['curSelection'] >= dlUL.children.length)
          {
            _AUTOCOMPLETE_DATA[elementID]['curSelection'] = -1;
            document.getElementById(elementID).value = _AUTOCOMPLETE_DATA[elementID]['curValue'];
            dlUL.scrollTop = 0;
          }
        }
        else
          _AUTOCOMPLETE_DATA[elementID]['curSelection'] = -1;
      }
      else if(event.keyCode == 27) // escape
      {
        closeDroplist();
      }
  // Backspace, delete, number and chars
      else if(event.keyCode == 8 ||
              event.keyCode == 46 ||
              (event.keyCode >= 48) ||
              (event.keyCode == 0 && event.key != ''))
      {
        var elem = document.getElementById(elementID);
        var value = elem.value;
        var dropdown, ddStyle;
        _AUTOCOMPLETE_DATA[elementID]['curValue'] = value;
        if(cbFunction == undefined)
        {
          if(value.length >= minLength)
          {
            if(document.getElementById(elementID).className.indexOf('ac_maininput') == -1)
              document.getElementById(elementID).className += ' ac_maininput';
            dropdown = document.getElementById('ac_' + elementID + '_dropdown');
            if(dropdown == null)
            {
              var rect;
              if(typeof(elem.getBoundingClientRect) == 'function')
                rect = elem.getBoundingClientRect();
              else
                rect = getElementRect(elem);
              var cs = document.defaultView.getComputedStyle(elem, null);
              ddStyle  = 'position: fixed;';
              ddStyle += 'min-width: ' + (rect.width - 4) + 'px; ';
              ddStyle += 'width: ' + (rect.width - 4) + 'px; ';
              ddStyle += 'top: ' + (rect.top + rect.height - 1) + 'px; ';
              ddStyle += 'left: ' + rect.left + 'px; ';
              ddStyle += 'border-top-color: ' + cs.getPropertyValue('border-bottom-color') + '; ';
              ddStyle += 'border-right-color: ' + cs.getPropertyValue('border-right-color') + '; ';
              dropdown = document.createElement('div');
              dropdown.setAttribute('id', 'ac_' + elementID + '_dropdown');
              dropdown.setAttribute('class', 'ac_dropdown');
              dropdown.setAttribute('style', ddStyle);
              document.body.appendChild(dropdown);
            }
            var key = '', otherParams = '';
            for(key in ajaxParams)
            {
              if(otherParams != '')
                otherParams += '&';
              if(typeof(ajaxParams[key]) == 'function')
                otherParams += key + '=' + ajaxParams[key]();
              else
                otherParams += key + '=' + ajaxParams[key];
            }
            if(otherParams != '')
              otherParams = '&' + otherParams;
            sendAJAXRequest(ajaxURL, searchKey + '=' + value + otherParams,
              function(response, status, params) {
                if(status == 200)
                {
                  var results = JSON.parse(response);
                  if(results.length != 0)
                  {
                    var extraStyle = '';
                    if(scrollAfter !== false)
                    {
                      extraStyle = 'style="max-height: ' + (scrollAfter * 20) + 'px; overflow: auto;"';
                    }
                    var list = '<ul class="ac_results" ' + extraStyle + '>';
                    for(var l = 0; l < results.length; l++)
                    {
                      var re = new RegExp('(' + value + ')', 'gi');
                      if(_AUTOCOMPLETE_DATA[elementID]['lineParser'] != undefined)
                      {
                        var complexResponse = _AUTOCOMPLETE_DATA[elementID]['lineParser'](results[l]);
                        var complexData = '';
                        for(key in complexResponse)
                        {
                          if(key != 'value')
                          {
                            if(complexData != '')
                              complexData += ',';
                            complexData += key + "=" + complexResponse[key];
                          }
                        }
                        list += '<li onMouseOver="Autocomplete.deleteHoverClass(\'' + elementID + '\');" data-complex="' + Base64.encode(complexData) + '" data-element="' + elementID + '" data-value="' + complexResponse['value'] + '">' + complexResponse['value'].replace(re, '<span class="ac_found">$1</span>') + '</li>';
                      }
                      else
                        list += '<li onMouseOver="Autocomplete.deleteHoverClass(\'' + elementID + '\');" data-element="' + elementID + '" data-value="' + results[l] + '">' + results[l].replace(re, '<span class="ac_found">$1</span>') + '</li>';
                    }
                    list += '</ul>';
                    params[0].innerHTML = list;
                    _AUTOCOMPLETE_DATA[elementID]['curSelection'] = -1;
                  }
                  else
                    closeDroplist();
                }
            }, new Array(dropdown));
          }
          else
          {
            closeDroplist();
          }
        }
        else
        {
          var key = '', otherParams = '';
          for(key in ajaxParams)
          {
            if(otherParams != '')
              otherParams += '&';
            if(typeof(ajaxParams[key]) == 'function')
              otherParams += key + '=' + ajaxParams[key]();
            else
              otherParams += key + '=' + ajaxParams[key];
          }
          if(otherParams != '')
            otherParams = '&' + otherParams;
          sendAJAXRequest(ajaxURL, searchKey + '=' + value + otherParams, cbFunction, cbParams);
        }
      }
    };
  };

  /**
   * Deletes the class "hover" from all LIs for the given elementID
   * Triggered on mouse over
   * 
   * @param string elementID                The element ID
   */
  var deleteHoverClass = function(elementID)
  {
    elems = document.getElementsByTagName('li');

    for(var l = 0; l < elems.length; l++)
    {
      if(elems[l].className == 'hover' && elems[l].getAttribute('data-element') == elementID)
        elems[l].className = '';
    }
  };

  document.addEventListener('click', function(e)
  {
    var elementID = '';

    e = e || window.event;
    if(e.which == 1) // only left click
    {
      var target = e.target || e.srcElement;
      var gotTarget = false;
      var targetValue = '';
      var elems;
      var dataComplex = '';
      while(target)
      {
        if(target.className == 'ac_dropdown')
          gotTarget = true;
        if(target.nodeName == 'LI')
        {
          targetValue = target.getAttribute('data-value');
          if(target.getAttribute('data-complex') !== null &&
              target.getAttribute('data-complex') != '')
            dataComplex = Base64.decode(target.getAttribute('data-complex'));
        }
        if(target.nodeName == 'INPUT')
        {
          if(_AUTOCOMPLETE_DATA.hasOwnProperty(target.id) && _AUTOCOMPLETE_DATA[target.id]['deleteOnClick'] == true)
            target.value = '';
        }
        target = target.parentElement;
      }
      if(gotTarget)
      {
        elems = document.getElementsByTagName('input');
        for(var l = 0; l < elems.length; l++)
        {
          if(elems[l].className.indexOf('ac_maininput') != -1)
          {
            elems[l].value = targetValue;
            elementID = elems[l].id;
          }
        }
        if(_AUTOCOMPLETE_DATA[elementID]['onSelect'] != undefined)
        {
          var userData = {};
          if(dataComplex !== null)
          {
            var data = dataComplex.split(',');
            for(l = 0; l < data.length; l++)
            {
              var x = data[l].split('=');
              userData[x[0]] = x[1];
            }
          }
          _AUTOCOMPLETE_DATA[elementID]['onSelect'](targetValue, userData);
        }
      }

      closeDroplist();
    }
  }, false);

  return { 'set': setAutocomplete,
           'getExtraJSParams' : getExtraJSParams,
           'deleteHoverClass' : deleteHoverClass,
           'version' : version };
}());

if(typeof(JSON) == 'undefined')
{
// Load JSON-parser for old browsers
  var scripts = document.getElementsByTagName('script');
  for(var s = 0; s < scripts.length; s++)
  {
    script = scripts[s];
    if(script.getAttribute.length !== undefined)
    {
      if(script.src.indexOf('autocomplete.js') != -1)
      {
        var imported = document.createElement('script');
        imported.src = script.src.replace('autocomplete.js', 'json_parse.js');
        document.getElementsByTagName('head')[0].appendChild(imported);
      }
    }
  }
}
if(typeof(Base64) == 'undefined')
{
// Load JSON-parser for old browsers
  var scripts = document.getElementsByTagName('script');
  for(var s = 0; s < scripts.length; s++)
  {
    script = scripts[s];
    if(script.getAttribute.length !== undefined)
    {
      if(script.src.indexOf('autocomplete.js') != -1)
      {
        var imported = document.createElement('script');
        imported.src = script.src.replace('autocomplete.js', 'base64.js');
        document.getElementsByTagName('head')[0].appendChild(imported);
      }
    }
  }
}
